/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <errno.h>
#include <glib.h>
#include <sys/mman.h>
#include <dlog.h>
#include <Elementary.h>
#include <Ecore_Wayland.h>
#include <wayland-extension/tizen-extension-client-protocol.h>
#include <tizen-remote-surface-client-protocol.h>
#include <tbm_surface.h>
#include <tbm_surface_internal.h>

#include "screen_connector_toolkit.h"
#include "screen_connector_toolkit_evas.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "SC_WATCHER_EVAS"

struct _screen_connector_type_evas_h {
	GHashTable *toolkit_table;
	Evas_Object *viewer_win;
};

struct _screen_connector_toolkit_evas_h {
	screen_connector_toolkit_evas_ops *ops;
	screen_connector_toolkit_h toolkit_h;
	int freeze;
	bool is_init;
	bool cancel_touch;
	Evas_Object *img_tbm;
	Evas_Object *img_file;
	uint32_t img_type;
	struct wl_buffer *pre_buffer;
	struct _screen_connector_type_evas_h *type_h;
	void *data;
};

static GHashTable *__type_table = NULL;
static Ecore_Event_Handler *__visibility_listener;

static void __destroy_type_h(gpointer data)
{
	screen_connector_type_evas_h type_h = (screen_connector_type_evas_h)data;
	if (!type_h)
		return;

	if (type_h->toolkit_table) {
		g_hash_table_destroy(type_h->toolkit_table);
		type_h->toolkit_table = NULL;
	}
	free(type_h);
}

EXPORT_API int screen_connector_toolkit_evas_init(Evas_Object *win, screen_connector_screen_type_e type)
{
	screen_connector_type_evas_h type_h;

	if (__type_table == NULL)
		__type_table = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, __destroy_type_h);

	type_h = g_hash_table_lookup(__type_table, GINT_TO_POINTER(type));
	if (type_h == NULL) {

		type_h = (screen_connector_type_evas_h)calloc(1, sizeof(struct _screen_connector_type_evas_h));
		g_hash_table_insert(__type_table, GINT_TO_POINTER(type), type_h);

		type_h->toolkit_table = g_hash_table_new(g_direct_hash, g_direct_equal);
		if (!type_h->toolkit_table) {
			LOGE("failed to create table");
			return -1;
		}
	}

	type_h->viewer_win = win;
	return screen_connector_toolkit_init(type);
}

EXPORT_API int screen_connector_toolkit_evas_fini(screen_connector_screen_type_e type)
{
	screen_connector_type_evas_h type_h;

	if (__type_table == NULL) {
		LOGI("__type_table is NULL");
		return 0;
	}

	type_h = g_hash_table_lookup(__type_table, GINT_TO_POINTER(type));
	if (type_h) {
		if (type_h->toolkit_table) {
			g_hash_table_destroy(type_h->toolkit_table);
			type_h->toolkit_table = NULL;
		}
		g_hash_table_remove(__type_table, GINT_TO_POINTER(type));
	}

	if (g_hash_table_size(__type_table) == 0) {
		g_hash_table_destroy(__type_table);
		__type_table = NULL;
	}

	return screen_connector_toolkit_fini(type);
}

static bool __obj_is_visible(Evas_Object *obj, screen_connector_toolkit_evas_h toolkit_evas_h)
{
	int x, y, w, h;
	Ecore_Wl_Window *window = NULL;
	int window_x, window_y, window_w, window_h;
	int rotation;

	if (toolkit_evas_h->type_h == NULL || toolkit_evas_h->type_h->viewer_win == NULL) {
		LOGE("type_h is NULL");
		return false;
	}

	window = elm_win_wl_window_get(toolkit_evas_h->type_h->viewer_win);
	ecore_wl_window_geometry_get(window, &window_x, &window_y, &window_w, &window_h);
	evas_object_geometry_get(obj, &x, &y, &w, &h);
	rotation = ecore_wl_window_rotation_get(window);

	if (x >= 0 && x < window_w &&
			y >= 0 && y < window_h &&
			(rotation == 0 || rotation == 180)) {
		LOGD("x %d, y %d w %d h %d, window_w %d window_h %d rotation %d",
				x, y, w, h, window_w, window_h, rotation);
		return true;
	} else if (x >= 0 && x < window_h &&
			y >= 0 && y < window_w &&
			(rotation == 90 || rotation == 270)) {
		LOGD("x %d, y %d w %d h %d, window_w %d window_h %d rotation %d",
				x, y, w, h, window_w, window_h, rotation);
		return true;
	}

	return false;
}

static int __set_visibility(screen_connector_toolkit_evas_h toolkit_evas_h, visibility_type type)
{
	int obscured;
	struct tizen_remote_surface *surface;

	if (!toolkit_evas_h) {
		LOGE("obj not found");
		return -1;
	}

	switch (type) {
	case VISIBILITY_TYPE_UNOBSCURED:
	case VISIBILITY_TYPE_PARTIALLY_OBSCURED:
		obscured = TIZEN_REMOTE_SURFACE_VISIBILITY_TYPE_VISIBLE;
		break;
	case VISIBILITY_TYPE_FULLY_OBSCURED:
		obscured = TIZEN_REMOTE_SURFACE_VISIBILITY_TYPE_INVISIBLE;
		break;
	default:
		return -1;
	}

	surface = screen_connector_toolkit_get_trs(toolkit_evas_h->toolkit_h);
	if (surface)
		tizen_remote_surface_transfer_visibility(surface, obscured);

	return 0;
}

static void __obj_update_visibility(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	screen_connector_toolkit_evas_h toolkit_evas_h = (screen_connector_toolkit_evas_h)data;

	if (!toolkit_evas_h->is_init) {
		toolkit_evas_h->is_init = true;
		return;
	}

	if (__obj_is_visible(obj, toolkit_evas_h))
		__set_visibility(toolkit_evas_h, VISIBILITY_TYPE_UNOBSCURED);
	else
		__set_visibility(toolkit_evas_h, VISIBILITY_TYPE_FULLY_OBSCURED);
}

static void __rs_cb_mouse_down(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	screen_connector_toolkit_evas_h toolkit_evas_h = (screen_connector_toolkit_evas_h)data;
	Evas_Event_Mouse_Down *ev = event_info;
	const char *desc = evas_device_description_get(ev->dev);
	int x;
	int y;
	int w;
	int h;
	struct tizen_remote_surface *surface;

	evas_object_geometry_get(obj, &x, &y, &w, &h);

	LOGD("mouse down: %d %d", ev->canvas.x - x, ev->canvas.y - y);

	if (desc == NULL)
		desc = "";

	if (screen_connector_toolkit_check_input_disabled(toolkit_evas_h->toolkit_h,
			SCREEN_CONNECTOR_INPUT_TYPE_MOUSE_UP_DOWN)) {
		LOGD("mouse down disabled");
		return;
	}

	surface = screen_connector_toolkit_get_trs(toolkit_evas_h->toolkit_h);
	if (surface == NULL) {
		LOGE("surface is NULL");
		return;
	}

	tizen_remote_surface_transfer_mouse_event(surface,
			TIZEN_REMOTE_SURFACE_EVENT_TYPE_MOUSE_DOWN,
			0,
			ev->button,
			ev->canvas.x - x,
			ev->canvas.y - y,
			wl_fixed_from_double(ev->radius_x),
			wl_fixed_from_double(ev->radius_y),
			wl_fixed_from_double(ev->pressure),
			wl_fixed_from_double(ev->angle),
			evas_device_class_get(ev->dev),
			evas_device_subclass_get(ev->dev),
			desc,
			ev->timestamp);
}

static void __rs_cb_mouse_up(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	screen_connector_toolkit_evas_h toolkit_evas_h = (screen_connector_toolkit_evas_h)data;
	Evas_Event_Mouse_Up *ev = event_info;
	const char *desc = evas_device_description_get(ev->dev);
	int x;
	int y;
	int w;
	int h;
	struct tizen_remote_surface *surface;

	evas_object_geometry_get(obj, &x, &y, &w, &h);

	LOGD("mouse up: %d %d", ev->canvas.x - x, ev->canvas.y - y);

	if (desc == NULL)
		desc = "";

	if (screen_connector_toolkit_check_input_disabled(toolkit_evas_h->toolkit_h,
			SCREEN_CONNECTOR_INPUT_TYPE_MOUSE_UP_DOWN)) {
		LOGD("mouse up disabled");
		return;
	}

	surface = screen_connector_toolkit_get_trs(toolkit_evas_h->toolkit_h);
	if (surface == NULL) {
		LOGE("surface is NULL");
		return;
	}

	tizen_remote_surface_transfer_mouse_event(surface,
			TIZEN_REMOTE_SURFACE_EVENT_TYPE_MOUSE_UP,
			0,
			ev->button,
			ev->canvas.x - x,
			ev->canvas.y - y,
			wl_fixed_from_double(ev->radius_x),
			wl_fixed_from_double(ev->radius_y),
			wl_fixed_from_double(ev->pressure),
			wl_fixed_from_double(ev->angle),
			evas_device_class_get(ev->dev),
			evas_device_subclass_get(ev->dev),
			desc,
			ev->timestamp);
}

static void __rs_cb_mouse_move(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	screen_connector_toolkit_evas_h toolkit_evas_h = (screen_connector_toolkit_evas_h)data;
	Evas_Event_Mouse_Move *ev = event_info;
	const char *desc = evas_device_description_get(ev->dev);
	int x;
	int y;
	int w;
	int h;
	struct tizen_remote_surface *surface;

	evas_object_geometry_get(obj, &x, &y, &w, &h);

	LOGD("mouse move: %d %d", ev->cur.canvas.x - x, ev->cur.canvas.y - y);

	if (desc == NULL)
		desc = "";

	if (screen_connector_toolkit_check_input_disabled(toolkit_evas_h->toolkit_h,
			SCREEN_CONNECTOR_INPUT_TYPE_MOUSE_MOVE_X_Y)) {
		LOGD("mouse move disabled");
		return;
	}

	surface = screen_connector_toolkit_get_trs(toolkit_evas_h->toolkit_h);
	if (surface == NULL) {
		LOGE("surface is NULL");
		return;
	}

	tizen_remote_surface_transfer_mouse_event(surface,
			TIZEN_REMOTE_SURFACE_EVENT_TYPE_MOUSE_MOVE,
			0,
			0,
			ev->cur.canvas.x - x,
			ev->cur.canvas.y - y,
			wl_fixed_from_double(ev->radius_x),
			wl_fixed_from_double(ev->radius_y),
			wl_fixed_from_double(ev->pressure),
			wl_fixed_from_double(ev->angle),
			evas_device_class_get(ev->dev),
			evas_device_subclass_get(ev->dev),
			desc,
			ev->timestamp);
}

static void __rs_cb_mouse_wheel(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	screen_connector_toolkit_evas_h toolkit_evas_h = (screen_connector_toolkit_evas_h)data;
	Evas_Event_Mouse_Wheel *ev = event_info;
	const char *desc = evas_device_description_get(ev->dev);
	struct tizen_remote_surface *surface;

	LOGD("mouse wheel");

	if (desc == NULL)
		desc = "";

	if (screen_connector_toolkit_check_input_disabled(toolkit_evas_h->toolkit_h,
			SCREEN_CONNECTOR_INPUT_TYPE_MOUSE_WHEEL)) {
		LOGD("mouse wheel disabled");
		return;
	}

	surface = screen_connector_toolkit_get_trs(toolkit_evas_h->toolkit_h);
	if (surface == NULL) {
		LOGE("surface is NULL");
		return;
	}

	tizen_remote_surface_transfer_mouse_wheel(surface,
			ev->direction,
			ev->z,
			evas_device_class_get(ev->dev),
			evas_device_subclass_get(ev->dev),
			desc,
			ev->timestamp);
}

static void __rs_cb_show(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	LOGD("show");
	__obj_update_visibility(data, e, obj, event_info);
}

static void __rs_cb_hide(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	LOGD("hide");
}

static void __rs_cb_resize(void *data, Evas *e, Evas_Object *obj, void *event_info)
{
	LOGD("resize");
	__obj_update_visibility(data, e, obj, event_info);
}

static void __clear_img_tbm(screen_connector_toolkit_evas_h toolkit_evas_h)
{
	if (toolkit_evas_h->img_tbm == NULL)
		return;

	g_hash_table_remove(toolkit_evas_h->type_h->toolkit_table, toolkit_evas_h->img_tbm);
	evas_object_del(toolkit_evas_h->img_tbm);
	toolkit_evas_h->img_tbm = NULL;
}

static void __clear_img_file(screen_connector_toolkit_evas_h toolkit_evas_h)
{
	if (toolkit_evas_h->img_file == NULL)
		return;

	evas_object_del(toolkit_evas_h->img_file);
	toolkit_evas_h->img_file = NULL;
}

static void __toolkit_update_cb(struct tizen_remote_surface *trs, uint32_t type, struct wl_buffer *tbm,
		int32_t img_file_fd, uint32_t img_file_size, uint32_t time, struct wl_array *keys,
		const char *appid, const char *instance_id, const int pid, void *data)
{
	bool is_added = false;
	Evas_Native_Surface ns;
	tbm_surface_h tbm_surface;
	int width, height;
	char *map;
	char format[] = "png";
	char plug_id[256];
	screen_connector_toolkit_evas_h toolkit_evas_h;
	tbm_surface_h pre_tbm_surface;

	toolkit_evas_h = (screen_connector_toolkit_evas_h)data;
	if (toolkit_evas_h->img_tbm == NULL)	{
		LOGD("first added !!!! %d, %d %s", (toolkit_evas_h->img_tbm == NULL), type, appid);
		toolkit_evas_h->img_tbm = evas_object_image_filled_add(
				evas_object_evas_get(toolkit_evas_h->type_h->viewer_win));

		/* Image alpha set */
		evas_object_image_alpha_set(toolkit_evas_h->img_tbm, 1);

		evas_object_event_callback_add(toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_MOUSE_DOWN,
				__rs_cb_mouse_down, toolkit_evas_h);
		evas_object_event_callback_add(toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_MOUSE_UP,
				__rs_cb_mouse_up, toolkit_evas_h);
		evas_object_event_callback_add(toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_MOUSE_MOVE,
				__rs_cb_mouse_move, toolkit_evas_h);
		evas_object_event_callback_add(toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_MOUSE_WHEEL,
				__rs_cb_mouse_wheel, toolkit_evas_h);
		evas_object_event_callback_add(toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_SHOW,
				__rs_cb_show, toolkit_evas_h);
		evas_object_event_callback_add(toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_HIDE,
				__rs_cb_hide, toolkit_evas_h);
		evas_object_event_callback_add(toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_RESIZE,
				__rs_cb_resize, toolkit_evas_h);

		/* Store toolkit_evas_h */
		g_hash_table_insert(toolkit_evas_h->type_h->toolkit_table, toolkit_evas_h->img_tbm, toolkit_evas_h);

		/* Set data to use in accessibility */
		snprintf(plug_id, sizeof(plug_id), "%s:%d", appid, pid);
		evas_object_data_set(toolkit_evas_h->img_tbm, "___PLUGID", strdup(plug_id));
		evas_object_event_callback_add(toolkit_evas_h->img_tbm,
				EVAS_CALLBACK_MOVE, __obj_update_visibility, toolkit_evas_h);
	}

	if (toolkit_evas_h->img_type != type)
		is_added = true;
	toolkit_evas_h->img_type = type;

	/* check type of given buffer */
	if (type == TIZEN_REMOTE_SURFACE_BUFFER_TYPE_TBM) {
		/* get tbm surface from buffer */
		tbm_surface = wl_buffer_get_user_data(tbm);
		width = tbm_surface_get_width(tbm_surface);
		height = tbm_surface_get_height(tbm_surface);

		memset(&ns, 0, sizeof(Evas_Native_Surface));
		ns.version = EVAS_NATIVE_SURFACE_VERSION;
		ns.type = EVAS_NATIVE_SURFACE_TBM;
		ns.data.tbm.buffer = tbm_surface;

		evas_object_resize(toolkit_evas_h->img_tbm, width, height);
		evas_object_image_size_set(toolkit_evas_h->img_tbm, width, height);
		evas_object_image_native_surface_set(toolkit_evas_h->img_tbm, &ns); /* set native surface */
		evas_object_image_pixels_dirty_set(toolkit_evas_h->img_tbm, EINA_TRUE); /* set dirty for image updating */
	} else if (type == TIZEN_REMOTE_SURFACE_BUFFER_TYPE_IMAGE_FILE) {
		if (!toolkit_evas_h->img_file)
			toolkit_evas_h->img_file = evas_object_image_filled_add(evas_object_evas_get(toolkit_evas_h->type_h->viewer_win));

		map = mmap(NULL, img_file_size, PROT_READ, MAP_SHARED, img_file_fd, 0);
		evas_object_image_memfile_set(toolkit_evas_h->img_file, map, img_file_size, format, NULL);
		evas_object_image_size_get(toolkit_evas_h->img_tbm, &width, &height);
		evas_object_image_fill_set(toolkit_evas_h->img_file, 0, 0, width, height);
		evas_object_resize(toolkit_evas_h->img_file, width, height);
		munmap(map, img_file_size);

		evas_object_image_native_surface_set(toolkit_evas_h->img_tbm, NULL); /* set null to previous object for the tbm type */
	}

	if (toolkit_evas_h->pre_buffer) {
		if (tizen_remote_surface_get_version(trs) >= TIZEN_REMOTE_SURFACE_RELEASE_SINCE_VERSION) {
			tizen_remote_surface_release(trs, toolkit_evas_h->pre_buffer);
			pre_tbm_surface = (tbm_surface_h)wl_buffer_get_user_data(toolkit_evas_h->pre_buffer);
			tbm_surface_internal_unref(pre_tbm_surface);
			wl_buffer_destroy(toolkit_evas_h->pre_buffer);
		}
	}
	toolkit_evas_h->pre_buffer = tbm;

	if (is_added) {
		LOGD("call added !!!! %s, type %d", appid, type);
		if (type == TIZEN_REMOTE_SURFACE_BUFFER_TYPE_IMAGE_FILE) {
			toolkit_evas_h->ops->added_cb(appid, instance_id, pid, toolkit_evas_h->img_file,
					toolkit_evas_h->data);
			__clear_img_tbm(toolkit_evas_h);
		} else {
			toolkit_evas_h->ops->added_cb(appid, instance_id, pid, toolkit_evas_h->img_tbm,
					toolkit_evas_h->data);
			__clear_img_file(toolkit_evas_h);
		}
	} else {
		LOGD("call updated !!!! %s, type %d", appid, type);
		if (type == TIZEN_REMOTE_SURFACE_BUFFER_TYPE_IMAGE_FILE) {
			toolkit_evas_h->ops->updated_cb(appid, instance_id, pid, toolkit_evas_h->img_file,
					toolkit_evas_h->data);
			__clear_img_tbm(toolkit_evas_h);
		} else {
			toolkit_evas_h->ops->updated_cb(appid, instance_id, pid, toolkit_evas_h->img_tbm,
					toolkit_evas_h->data);
			__clear_img_file(toolkit_evas_h);
		}
	}
	close(img_file_fd); /* close passed fd */
}

static void __toolkit_added_cb(const char *appid, const char *instance_id, const int pid, void *data)
{
}

static void __free_toolkit_evas_h(screen_connector_toolkit_evas_h toolkit_evas_h)
{
	if (!toolkit_evas_h)
		return;

	if (toolkit_evas_h->toolkit_h)
		screen_connector_toolkit_remove(toolkit_evas_h->toolkit_h);

	__clear_img_file(toolkit_evas_h);
	__clear_img_tbm(toolkit_evas_h);

	if (toolkit_evas_h->ops) {
		free(toolkit_evas_h->ops);
		toolkit_evas_h->ops = NULL;
	}

	free(toolkit_evas_h);
}

static void __toolkit_removed_cb(const char *appid, const char *instance_id, const int pid, void *data)
{
	screen_connector_toolkit_evas_h toolkit_evas_h;
	char *plug_id;

	LOGD("missing %s %s", appid, instance_id);
	toolkit_evas_h = (screen_connector_toolkit_evas_h)data;
	if (!toolkit_evas_h) {
		LOGE("Null toolkit_evas_h");
		return;
	}

	if (toolkit_evas_h->type_h->toolkit_table != NULL && toolkit_evas_h->img_tbm != NULL) {
		g_hash_table_remove(toolkit_evas_h->type_h->toolkit_table, toolkit_evas_h->img_tbm);

		/* Remove data used in accessibility */
		plug_id = evas_object_data_del(toolkit_evas_h->img_tbm, "___PLUGID");
		free(plug_id);
	}

	if (toolkit_evas_h->img_tbm != NULL)
		toolkit_evas_h->ops->removed_cb(appid, instance_id, pid, toolkit_evas_h->img_tbm, toolkit_evas_h->data);
	else if (toolkit_evas_h->img_file != NULL)
		toolkit_evas_h->ops->removed_cb(appid, instance_id, pid, toolkit_evas_h->img_file, toolkit_evas_h->data);

	__free_toolkit_evas_h(toolkit_evas_h);
}

EXPORT_API screen_connector_toolkit_evas_h screen_connector_toolkit_evas_add(screen_connector_toolkit_evas_ops *ops, char *id,
		screen_connector_screen_type_e type, void *data)
{
	screen_connector_toolkit_evas_h handle;
	screen_connector_toolkit_evas_ops *evas_ops;
	screen_connector_toolkit_ops toolkit_ops;
	screen_connector_type_evas_h type_h;
	int int_type = type;

	type_h = g_hash_table_lookup(__type_table, GINT_TO_POINTER(int_type));
	if (type_h == NULL) {
		LOGE("type %d is not initialized !!", type);
		return NULL;
	}

	toolkit_ops.added_cb = __toolkit_added_cb;
	toolkit_ops.updated_cb = __toolkit_update_cb;
	toolkit_ops.removed_cb = __toolkit_removed_cb;

	evas_ops = (screen_connector_toolkit_evas_ops *)calloc(1, sizeof(screen_connector_toolkit_evas_ops));
	if (evas_ops == NULL) {
		LOGE("evas_ops calloc fail");
		return NULL;
	}

	memcpy(evas_ops, ops, sizeof(screen_connector_toolkit_evas_ops));
	handle = (screen_connector_toolkit_evas_h)calloc(1, sizeof(struct _screen_connector_toolkit_evas_h));
	if (handle == NULL) {
		LOGE("handle calloc fail");
		free(evas_ops);
		return NULL;
	}
	handle->img_type = -1;
	handle->ops = evas_ops;
	handle->data = data;
	handle->type_h = type_h;
	handle->toolkit_h = screen_connector_toolkit_add(&toolkit_ops, id, type, handle);

	return handle;
}

EXPORT_API int screen_connector_toolkit_evas_remove(screen_connector_toolkit_evas_h handle)
{
	if (handle == NULL)
		return -1;

	__free_toolkit_evas_h(handle);
	return 0;
}

EXPORT_API int screen_connector_toolkit_evas_update(const char *appid, const char *instance_id)
{
	return screen_connector_toolkit_update(appid, instance_id);
}

static void __send_visibility(gpointer key, gpointer value, gpointer user_data)
{
	screen_connector_toolkit_evas_h toolkit_evas_h = (screen_connector_toolkit_evas_h)value;
	Evas_Object *evas_obj = (Evas_Object *)key;
	unsigned int event = GPOINTER_TO_INT(user_data);
	int ret;
	visibility_type type;

	if (toolkit_evas_h->freeze)
		return;

	if (!__obj_is_visible(evas_obj, toolkit_evas_h))
		return;

	if (event)
		type = VISIBILITY_TYPE_FULLY_OBSCURED;
	else
		type = VISIBILITY_TYPE_UNOBSCURED;

	ret = __set_visibility(toolkit_evas_h, type);
	if (ret == -1)
		LOGE("failed to set object visibility  set %p to %d", evas_obj, type);
}

static Eina_Bool __visibility_cb(void *data, int type, void *event)
{
	Ecore_Wl_Event_Window_Visibility_Change *ev = event;
	GHashTableIter iter;
	gpointer key, value;
	screen_connector_type_evas_h cur_type;

	LOGD("visibility change: %d %d", (unsigned int)ev->win,
			(unsigned int)ev->fully_obscured);

	if (!__type_table)
		return ECORE_CALLBACK_RENEW;

	g_hash_table_iter_init(&iter, __type_table);
	while (g_hash_table_iter_next(&iter, &key, &value)) {
		cur_type = (screen_connector_type_evas_h)value;
		if (cur_type->toolkit_table)
			g_hash_table_foreach(cur_type->toolkit_table, __send_visibility,
					GINT_TO_POINTER(ev->fully_obscured));
	}

	return ECORE_CALLBACK_RENEW;
}

EXPORT_API int screen_connector_toolkit_evas_start_visibility_notify(void)
{
	if (__visibility_listener)
		return 0;

	__visibility_listener = ecore_event_handler_add(ECORE_WL_EVENT_WINDOW_VISIBILITY_CHANGE, __visibility_cb, NULL);

	return 0;
}

EXPORT_API int screen_connector_toolkit_evas_stop_visibility_notify(void)
{
	if (!__visibility_listener)
		return 0;

	ecore_event_handler_del(__visibility_listener);
	__visibility_listener = NULL;

	return 0;
}

screen_connector_toolkit_evas_h __find_toolkit_evas_h(Evas_Object *obj)
{
	screen_connector_toolkit_evas_h toolkit_evas_h;
	GHashTableIter iter;
	gpointer key, value;
	screen_connector_type_evas_h cur_type;

	g_hash_table_iter_init(&iter, __type_table);
	while (g_hash_table_iter_next(&iter, &key, &value)) {
		cur_type = (screen_connector_type_evas_h)value;
		if (cur_type->toolkit_table) {
			toolkit_evas_h = g_hash_table_lookup(cur_type->toolkit_table, obj);
			if (toolkit_evas_h != NULL)
				return toolkit_evas_h;
		}
	}

	return NULL;
}

EXPORT_API int screen_connector_toolkit_evas_freeze_visibility(Evas_Object *obj, visibility_type type)
{
	screen_connector_toolkit_evas_h toolkit_evas_h;

	toolkit_evas_h = __find_toolkit_evas_h(obj);
	if (!toolkit_evas_h) {
		LOGE("obj not found");
		return -1;
	}
	toolkit_evas_h->freeze = 1;

	return __set_visibility(toolkit_evas_h, type);
}

EXPORT_API int screen_connector_toolkit_evas_thaw_visibility(Evas_Object *obj)
{
	screen_connector_toolkit_evas_h toolkit_evas_h;

	toolkit_evas_h = __find_toolkit_evas_h(obj);
	if (!toolkit_evas_h) {
		LOGE("obj not found");
		return -1;
	}
	toolkit_evas_h->freeze = 0;

	return 0;
}

EXPORT_API int screen_connector_toolkit_evas_send_mouse_up(Evas_Object *obj)
{
	screen_connector_toolkit_evas_h toolkit_evas_h;
	int x;
	int y;
	int w;
	int h;
	unsigned int timestamp;
	struct tizen_remote_surface *surface;

	toolkit_evas_h = __find_toolkit_evas_h(obj);
	if (!toolkit_evas_h) {
		LOGE("unknown object. not widget object.");
		return -1;
	}

	if (!toolkit_evas_h->img_tbm) {
		LOGE("surface not yet initialized.");
		return -1;
	}

	surface = screen_connector_toolkit_get_trs(toolkit_evas_h->toolkit_h);
	if (surface == NULL) {
		LOGE("surface is NULL");
		return -1;
	}

	evas_object_geometry_get(obj, &x, &y, &w, &h);
	timestamp = (unsigned int)((unsigned long long)(ecore_time_get() * 1000.0) & 0xffffffff);
	tizen_remote_surface_transfer_mouse_event(surface,
			TIZEN_REMOTE_SURFACE_EVENT_TYPE_MOUSE_UP,
			0,
			1,
			w / 2,
			h / 2,
			0,
			0,
			0,
			0,
			TIZEN_INPUT_DEVICE_CLAS_MOUSE,
			TIZEN_INPUT_DEVICE_SUBCLAS_NONE,
			"",
			timestamp);

	return 0;
}

EXPORT_API int screen_connector_toolkit_evas_send_touch_cancel(Evas_Object *obj)
{
	screen_connector_toolkit_evas_h toolkit_evas_h;
	struct tizen_remote_surface *surface;

	toolkit_evas_h = __find_toolkit_evas_h(obj);
	if (!toolkit_evas_h) {
		LOGE("unknown object. not widget object.");
		return -1;
	}

	if (toolkit_evas_h->img_tbm) {
		surface = screen_connector_toolkit_get_trs(toolkit_evas_h->toolkit_h);
		if (surface == NULL) {
			LOGE("surface is NULL");
			return -1;
		}

		tizen_remote_surface_transfer_touch_cancel(surface);
		toolkit_evas_h->cancel_touch = true;
	} else {
		LOGE("surface not yet initialized.");
		return -1;
	}

	return 0;
}

EXPORT_API screen_connector_toolkit_evas_h screen_connector_toolkit_evas_add_by_rid(screen_connector_toolkit_evas_ops *ops,
	int res_id, screen_connector_screen_type_e type, void *data)
{
	screen_connector_toolkit_evas_h handle;
	screen_connector_toolkit_evas_ops *evas_ops;
	screen_connector_toolkit_ops toolkit_ops;
	screen_connector_type_evas_h type_h;
	char rid_str[32];

	type_h = g_hash_table_lookup(__type_table, GINT_TO_POINTER(type));
	if (type_h == NULL) {
		LOGE("type %d is not initialized !!", type);
		return NULL;
	}

	toolkit_ops.added_cb = __toolkit_added_cb;
	toolkit_ops.updated_cb = __toolkit_update_cb;
	toolkit_ops.removed_cb = __toolkit_removed_cb;

	evas_ops = (screen_connector_toolkit_evas_ops *)calloc(1, sizeof(screen_connector_toolkit_evas_ops));
	if (evas_ops == NULL) {
		LOGE("evas_ops calloc fail");
		return NULL;
	}

	memcpy(evas_ops, ops, sizeof(screen_connector_toolkit_evas_ops));
	handle = (screen_connector_toolkit_evas_h)calloc(1, sizeof(struct _screen_connector_toolkit_evas_h));
	if (handle == NULL) {
		LOGE("handle calloc fail");
		free(evas_ops);
		return NULL;
	}
	handle->img_type = -1;
	handle->ops = evas_ops;
	handle->data = data;
	handle->type_h = type_h;

	snprintf(rid_str, sizeof(rid_str), "%d", res_id);
	handle->toolkit_h = screen_connector_toolkit_create_handle(rid_str, res_id, &toolkit_ops, handle);
	screen_connector_toolkit_redirect_surface(handle->toolkit_h);

	return handle;
}

EXPORT_API int screen_connector_toolkit_evas_get_rid(Evas_Object *obj, int *resource_id)
{
	screen_connector_toolkit_evas_h toolkit_evas_h;

	toolkit_evas_h = __find_toolkit_evas_h(obj);
	if (!toolkit_evas_h) {
		LOGE("unknown object.");
		return -1;
	}
	*resource_id = screen_connector_toolkit_get_surface_id(toolkit_evas_h->toolkit_h);

	return 0;
}

EXPORT_API int screen_connector_toolkit_evas_get_pid(Evas_Object *obj, int *pid)
{
	screen_connector_toolkit_evas_h toolkit_evas_h;
	int toolkit_pid = 0;

	toolkit_evas_h = __find_toolkit_evas_h(obj);
	if (!toolkit_evas_h) {
		LOGE("unknown object.");
		return -1;
	}

	toolkit_pid = screen_connector_toolkit_get_pid(toolkit_evas_h->toolkit_h);
	if (toolkit_pid == 0) {
		LOGE("Invalid object");
		return -1;
	}

	*pid = toolkit_pid;

	return 0;
}
